% Module infer
%
% Copyright (C) 2003 Alwen Tiu
%
%   This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or (at
% your option) any later version.
% 
%   This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software Foundation,
% Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



% Module: infer
% Author: Alwen Tiu
% Date: 8/26/03
% Description:
%   - contain main definition for inference schemes
%   - there are two predicates for the definition of inference rules:
%      inf and inff.
%     The predicate inff is a special case of inf, which is used when
%     user input (e.g. for instantiating sigma-r rule or invariants)
%     is obtained from a file instead of standard input.
%
%
% Revision:
% 8/31/03: - a mistake found in makeInstClause. Unbound variable G1 fixed.
%          - extend focus with equality-left.

module infer.

accum_sig unif, lincparser.
accumulate control, list, stt, unif, lincparser, tokenizer, printer.


load_def File Def :-
   if (open_in File In) (close_in In)
     (print "Failed in opening file ", print File, print "\n", fail),
   parsefile File (linc_root R),
   absyn_to_def R Def, !,
   if (defOK Def) (print "Definition loaded!\n") 
      (print "Definition not well-typed!\n", fail).

load_seq File LSeq N :-
   parsefile File (linc_root R),
   absyn_to_seq R LSeq', !,
   current_def (def Nm Kn Sig Cls), !,
   print "Parsing and abstract syntax translation OK\n",
   seqListOK Kn Sig LSeq' N, !,
   print "Type checking OK\n",
   norm_ListSeq Sig LSeq' LSeq.

   
inf inf_init G (seq Nm LFv Hyp G)  nil 0.
   
inf inf_top ltrue (seq Nm LFv Hyp ltrue) nil 0.

inf inf_bot lfalse (seq Nm LFv Hyp G) nil 0.

inf inf_eqr G (seq Nm LFv Hyp G) nil 0 :- G = (leq @ T @ T).
         
inf inf_andr G (seq Nm LFv Hyp G) LSeq 2 :-
   G = (land @ A @ B),
   Nm1 is Nm ^ "1",
   Nm2 is Nm ^ "2",
   LSeq = [(seq Nm1 LFv Hyp A), (seq Nm2 LFv Hyp B)].
   
inf inf_orr1 G (seq Nm LFv Hyp G) LSeq 1 :-
   G = (lor @ A @ B),
   Nm1 is Nm ^ "1",
   LSeq = [(seq Nm1 LFv Hyp A)].

inf inf_orr2 G (seq Nm LFv Hyp G) LSeq 1 :-
   G = (lor @ A @ B),
   Nm1 is Nm ^ "1",
   LSeq = [(seq Nm1 LFv Hyp B)].
         
inf inf_impr G (seq Nm LFv Hyp G) LSeq 1 :-
   G = (limp @ A @ B),
   Nm1 is Nm ^ "1",
   LSeq = [(seq Nm1 LFv [A | Hyp] B)].

inf inf_sigmar G (seq Nm LFv Hyp G) LSeq 1 :-
   G = (lexists @ L),
   Nm1 is Nm ^ "1",
   print "\nEnter expression: ",
   input_term T,
   current_def (def DefNm Kind Sig LCls),
   Seq' = (seq Nm1 LFv Hyp (L @ T)),
   seqOK Kind Sig Seq', norm_seq Sig Seq' Seq1,
   LSeq = [Seq1].

inf inf_pir G (seq Nm LFv Hyp G) LSeq 1 :-
   G = (lall @ L),
   Nm1 is Nm ^ "1",
   current_def (def _ Kn Sig _),
   append Sig LFv Sig',
   typeof Kn Sig' L (Ty1 --> Ty2), 
   get_next_fvar Sig' 0 V, !,
   eval [sg V Ty1|Sig'] (L @ V) G1,
   LSeq = [seq Nm1 [sg V Ty1 | LFv] Hyp G1].
   
inf inf_defr G (seq Nm LFv Hyp G) LSeq 1 :-
   atomic G [A|Args],
   Nm1 is Nm ^ "1",
   current_def (def DefNm Kind Sig LCls),
   member (defcl _ A B) LCls, !,
   apply_atom B Args C,
   append LFv Sig Sig',
   eval Sig' C D,         
   LSeq = [seq Nm1 LFv Hyp D].

inf inf_coindr G (seq Nm LFv Hyp G) LSeq 2 :-
   Nm1 is Nm ^ "1", Nm2 is Nm ^ "2",
   atomic G [A|Args],
   current_def (def DefNm Kind Sig LCls),
   member (defcl coindcl A B) LCls, !,
   print "\nEnter invariant: ",
   input_term I,  % input the induction invariant
   append LFv Sig Sig', 
   typeof Kind Sig' I Ty, typeof Kind Sig' A Ty,
   subst_invariant A I B B', 
   new_var_args 0 Args Kind Sig' NewSig NewVars,  % create new eigenvariables 
   apply_atom I Args C,   
   apply_atom B' NewVars BI,
   apply_atom I NewVars SI,
   eval Sig' C D,         
   append NewSig Sig Sig2,
   eval Sig2 BI BInorm,
   eval Sig2 SI SInorm,
   LSeq = [seq Nm1 NewSig [SInorm] BInorm, seq Nm1 LFv Hyp D].
   
inf inf_contrl H (seq Nm LFv Hyp G) LSeq 1 :-
   Nm1 is Nm ^ "1",
   LSeq = [seq Nm1 LFv [H,H|Hyp] G].

inf inf_weakl F (seq Nm LFv Hyp G) LSeq 1 :-
   Nm1 is Nm ^ "1",
   LSeq = [seq Nm1 LFv Hyp G].
            
inf inf_andl (land @ R @ T) (seq Nm LFv Hyp G) LSeq 1 :-
   Nm1 is Nm ^ "1",
   LSeq = [seq Nm1 LFv [R,T|Hyp] G].

inf inf_orl (lor @ R @ T) (seq Nm LFv Hyp G) LSeq 2 :-
   Nm1 is Nm ^ "1", Nm2 is Nm ^ "2",
   LSeq = [(seq Nm1 LFv [R | Hyp] G), (seq Nm2 LFv [T | Hyp] G) ].

inf inf_impl (limp @ R @ T) (seq Nm LFv Hyp G) LSeq 2 :-
   Nm1 is Nm ^ "1", Nm2 is Nm ^ "2",
   LSeq = [seq Nm1 LFv Hyp R, seq Nm2 LFv [T|Hyp] G].

inf inf_sigmal (lexists @ L) (seq Nm LFv Hyp G) LSeq 1 :-
   Nm1 is Nm ^ "1",
   current_def (def _ Kn Sig _),
   append Sig LFv Sig',
   typeof Kn Sig' L (Ty1 --> Ty2), 
   get_next_fvar Sig' 0 V, !,
   eval [sg V Ty1| Sig'] (L @ V) H1,
   LSeq = [seq Nm1 [sg V Ty1|LFv] [H1|Hyp] G].

inf inf_pil (lall @ L) (seq Nm LFv Hyp G) LSeq 1 :-
   Nm1 is Nm ^ "1",
   print "\nEnter expression: ",
   input_term T,
   current_def (def DefNm Kind Sig LCls),
   Seq' = (seq Nm1 LFv [L @ T|Hyp] G),
   seqOK Kind Sig Seq', norm_seq Sig Seq' Seq1,
   LSeq = [Seq1].
            
inf inf_eql (leq @ S @ T) (seq Nm LFv Hyp G) LSeq N :-
   current_def Def,
   unify Def LFv LFv1 S T Node LSubs,
   if (Node = succ_node)
      (
         print "Unification succeeded! List of unifiers: \n",
         print_list_subst LSubs 1,
         print "New signature : \n",
         print_signat std_out LFv1,
         print "\n",
         print "\nApplying unifiers to sequent\n",
         subst_seq LFv1 LSubs (seq Nm LFv Hyp G) LSeq N
      )
      ( 
         if (Node = flex_node) 
            (
               print "\n Flex-flex pairs found! Unification was not performed.\n",
               fail            
            )
            (  
               print "Unification failed.\n",
               LSeq = nil, N = 0
            )
      ).

inf inf_defl P (seq Nm LFv Hyp G) LSeq 1 :-
   atomic P [A|Args],
   Nm1 is Nm ^ "1",
   current_def (def DefNm Kind Sig LCls),
   member (defcl _ A B) LCls, !,
   apply_atom B Args C,
   append LFv Sig Sig',
   eval Sig' C D,         
   LSeq = [seq Nm1 LFv [D|Hyp] G].

inf inf_indl P (seq Nm LFv Hyp G) LSeq 2 :- 
   Nm1 is Nm ^ "1", Nm2 is Nm ^ "2",
   atomic P [A|Args],
   current_def (def DefNm Kind Sig LCls),
   member (defcl indcl A B) LCls, !,
   print "\nEnter invariant: ",
   input_term I,  % input the induction invariant
   append LFv Sig Sig', 
   typeof Kind Sig' I Ty, typeof Kind Sig' A Ty,
   subst_invariant A I B B', 
   new_var_args 0 Args Kind Sig' NewSig NewVars,  % create new eigenvariables 
   apply_atom I Args C,   
   apply_atom B' NewVars BI,
   apply_atom I NewVars SI,
   eval Sig' C D,         
   append NewSig Sig Sig2,
   eval Sig2 BI BInorm,
   eval Sig2 SI SInorm,
   LSeq = [seq Nm1 NewSig [BInorm] SInorm, seq Nm1 LFv [D|Hyp] G].
   
% -- special rules

type get_defr_clauses sequent -> list tm -> o.

get_defr_clauses (seq Nm LFv Hyp G) LB :-
	atomic G [A|Args],
   	current_def (def DefNm Kind Sig LCls),
   	member (defcl _ A B) LCls, !,
   	apply_atom B Args C,
   	append LFv Sig Sig',
   	eval Sig' C D,
   	getClauses D LB, !.
	
inf inf_defr_one  H (seq Nm LFv Hyp G) LSeq N :-
   ((pi X\ constant X :- member X LFv) => instClauses [H] LInst S),
   !,
   if (S = "ok")
      (
        if (makeInstSeq LInst (seq Nm LFv Hyp G) Seq)
           ( LSeq = [Seq], N = 1)
           ( LSeq = nil, N = 0),
        print "\nInstantiation ok\n"        
      )
      (fail).

inf inf_defr_inst G (seq Nm LFv Hyp G) LSeq N :-
   atomic G [A|Args],
   Nm1 is Nm ^ "1",
   current_def (def DefNm Kind Sig LCls),
   member (defcl _ A B) LCls, !,
   apply_atom B Args C,
   append LFv Sig Sig',
   eval Sig' C D,
   getClauses D LB, !,
   print "\nList of clauses: \n",
   print_listTerms std_out LB,
   print "Instantiating clauses...\n",
   ((pi X\ constant X :- member X LFv) => instClauses LB LInstB SR),   
   if (SR = "ok")
      ( 
%        print "\nList of terms:\n",
%        print_listTerms std_out LInstB,
%        print "\n",
        if (makeInstSeq LInstB (seq Nm LFv Hyp G) Seq)
           (LSeq = [Seq], N = 1)
           (LSeq = nil, N = 0),
        print "\nInstantiation successful\n"
      )
      ( fail).  
   

% --- helper predicates for instantiating def-r

% make a disjunction of conjunction 

type makeInstSeq list (list tm) -> sequent -> sequent -> o.

makeInstSeq [L | R] Seq Seq' :-   
   if (makeInstConj L Seq Seq1)
      (
         Seq1 = (seq Nm LFv Hyp G1),
         if (makeInstSeq R Seq Seq2)
            (
               Seq2 = (seq Nm LFv Hyp G2),
               Seq' = (seq Nm LFv Hyp (lor @ G1 @ G2))
            )
            (
               Seq' = Seq1
            )
      )
      (
         makeInstSeq R Seq Seq'
      ).

% make a conjunction of the formula in a list      
type makeInstConj list tm -> sequent -> sequent -> o.

makeInstConj [T|L] (seq Nm LFv Hyp G) Seq :-
   if (makeInstConj L (seq Nm LFv Hyp G) Seq')
      (
         Seq' = (seq Nm LFv Hyp G1),
         Seq = (seq Nm LFv Hyp (land @ T @ G1))
      )
      ( Seq = (seq Nm LFv Hyp T)).
   
   
% get the list of disjunction
type getClauses tm -> list tm -> o.

getClauses (lor @ A @ B) L :-
   !, getClauses A L1, !, getClauses B L2, !,
   append L1 L2 L.
getClauses X [X] :- !.

type getEquations tm -> list tm -> list tm -> o.

getEquations (leq @ A @ B) [leq @ A @ B] nil :- !.
getEquations (land @ A @ B) L R :-
  !, getEquations A L1 R1, !, getEquations B L2 R2, !,
  append L1 L2 L, append R1 R2 R.
getEquations X nil [X] :- !.  


   

type instOneClause list signat -> tm -> list tm -> o.

instOneClause SigAcc (lexists @ L) LTm :-
   !, current_def (def Nm Kn SigDef LCls),
   typeof Kn SigAcc L (Ty1 --> fo),
   get_next_fvar SigAcc 0 V,
   Sg = (sg V Ty1),
   eval [Sg|SigAcc] (L @ V) R,
   instOneClause [Sg|SigAcc] R LTm.
instOneClause SigAcc X LTm :-
   getEquations X LEq LR, !,
   print "\nBody of definition:\n",
   print_lterm std_out 0 X, print "\n",
   solveEquations SigAcc LEq LR LTm.

type lsubst_list_tm list signat -> list (list spair) -> list tm -> list tm -> o.

lsubst_list_tm Sig nil L nil.
lsubst_list_tm Sig [Subs | LSubs] L R :-
   subst_list_tm Sig Subs L L1,
   norm_terms Sig L1 L2,
   lsubst_list_tm Sig LSubs L R1,
   append L2 R1 R.
   
type instClauses list tm -> list (list tm) -> string -> o.

instClauses nil nil "fail".
instClauses [T|L] R S :-
   if (instOneClause nil T L1) 
      (instClauses L R1 S', 
       %append L1 R1 R, 
       R = [L1 | R1],
       S = "ok")
      (instClauses L R S).

type solveEquations list signat -> list tm -> list tm -> list tm -> o.
   
solveEquations Sig nil L L.
solveEquations Sig [(leq @ S @ T) | L] LR LTm :-
   current_def Def,
%   print "\nSolving equation: ",
%   print_lterm std_out 0 (leq @ S @ T), print "\n",
   unify Def Sig SigOut S T Node LSubs,
   if (Node = succ_node ; Node = flex_node)
      (
         lsubst_list_tm SigOut LSubs L L',
         lsubst_list_tm SigOut LSubs LR LR',
         solveEquations SigOut L' LR' LTm
      )
      (fail).
           
% ---
   
type input_term tm -> o.

input_term T :-
   input_line std_in S,
   N is (size S), N' is N - 1,
   S1 is (substring S 0 N'),
   S2 is "(" ^ S1 ^ ")\n",
   open_out "~temp.seq" Out,
   output Out S2,
   close_out Out,
   open_in "~temp.seq" In,
   ((pi It\ (pi LA\ (pi (LN\ (yyinitial LN It [eols] :- 
	 	   lookahead It LA, eof_symbol LA, !))))) => (eof_symbol "\n" => yyinitial 0 In Tokens)),
   close_in In, !,
   startparse (linc_term St) [bols] Tokens D Str,
   D = (linc_term T), !.
   
   
   
type apply_atom tm -> list tm -> tm -> o.

apply_atom M nil M.
apply_atom M [N|L] P :-
   apply_atom (M @ N) L P.
   
   
atomic (fvar S) [fvar S] :- !.
atomic (M @ N) L :- 
   !, atomic M L1, append L1 [N] L.
% atomic X [X] :- constant (sg X _).


type subst_invariant tm -> tm -> tm -> tm -> o.

subst_invariant P I B D :-
   (linc_copy P I :- !) => linc_copy B D.
      
   
% --- applying substitution to sequents

type subst_seq list signat -> list (list spair) -> sequent -> list sequent -> int -> o.
type subst_list_tm list signat -> list spair -> list tm -> list tm -> o.

type subst_sig list spair -> list signat -> list signat -> o.

subst_sig Subst nil nil.
subst_sig Subst [(sg V Ty) | L] R :-
   fvar_in_domain V Subst, not(fvar_in_range V Subst), !,
   subst_sig Subst L R.
subst_sig Subst [(sg V Ty) | L] [(sg V Ty) | R] :-
   subst_sig Subst L R.


type fvar_in_domain tm -> list spair -> o.
fvar_in_domain V [(pair V _) | L] :- !.
fvar_in_domain V [_ | L] :- fvar_in_domain V L.

type fvar_in_term tm -> tm -> o.

fvar_in_term V V :- !.
fvar_in_term V (M @ N) :- fvar_in_term V M; fvar_in_term V N.
fvar_in_term V (lam Ty L) :- pi x\ fvar_in_term V (L x).

type fvar_in_range tm -> list spair -> o.
fvar_in_range V [(pair _ T) | L] :- fvar_in_term V T, !.
fvar_in_range V [_ | L] :- fvar_in_range V L.



subst_list_tm Sig Subs nil nil.
subst_list_tm Sig Subs [T | L] [T' | L'] :-
   subst Sig Subs T T',
   subst_list_tm Sig Subs L L'.

subst_seq Sig nil (seq Nm Sig' Hyp G) nil 0.
subst_seq Sig [Subs | L] (seq Nm Sig' Hyp G) LSeq N :-
   subst_list_tm Sig Subs [G|Hyp] [G'|Hyp'],
   Seq' = (seq Nm Sig Hyp' G'),
   current_def (def _ _ SigDef _),
   % normalize the sequent and trim the signature
%   print "\nSequent before normalization:\n",
%   print_sequent std_out Seq',
   norm_seq SigDef Seq' (seq Nm Sig Hyp1 G1),
%   print "\nApplying substitution to signature:\n",
%   print_signat std_out Sig,
   subst_sig Subs Sig Sig2,      
%   print "\nThe resulting signature:\n",
%   print_signat std_out Sig2,
%   print "\n",
   Seq2 = (seq Nm Sig2 Hyp1 G1),
%   print "\nThe resulting sequent: \n",
%   print_sequent std_out Seq2,
%   print "\n",
   subst_seq Sig L (seq Nm Sig' Hyp G) TSeq N',
   LSeq = [Seq2 | TSeq], N is N' + 1.

   
type print_subst1 list spair -> o.
type print_subst list spair -> o.
type print_list_subst list (list spair) -> int -> o.

print_subst1 nil.
print_subst1 [(pair V T) | L] :-
   print_lterm std_out 0 V,  print " |-> ", print_lterm std_out 0 T,
   if (L = nil) (true) (print ", ", print_subst1 L).

print_subst L :- print "[", print_subst1 L, print "]".

print_list_subst nil N.
print_list_subst [S|L] N :-
  print "\nUnifier ", term_to_string N Str, print Str,
  print ":\n", print_subst S, N' is N + 1, print "\n",
  print_list_subst L N'.
     
  
% --- create new eigenvariables for induction
type new_var_args int -> list tm -> list ty -> list signat -> list signat -> list tm -> o.

new_var_args N nil Kind Sig nil nil.
new_var_args N [T|L] Kind Sig [(sg V Ty) | R] [V | LV] :-
   typeof Kind Sig T Ty,
   term_to_string N NStr,
   Str is "Z_" ^ NStr, V = (fvar Str),
   N' is N + 1,
   new_var_args N' L Kind Sig R LV.
   


type norm_terms list signat -> list tm -> list tm -> o.


norm_terms Sig nil nil.
norm_terms Sig [H | L] [H' | L'] :-
   eval Sig H H', norm_terms Sig L L'.
   
norm_seq Sig1 (seq Nm Sig2 Hyp G) (seq Nm Sig2 Hyp' G') :-
   append Sig1 Sig2 Sig,
   eval Sig G G', 
   norm_terms Sig Hyp Hyp'.

type norm_ListSeq list signat -> list sequent -> list sequent -> o.

norm_ListSeq Sig nil nil.
norm_ListSeq Sig [Seq | L] [Seq' | L'] :-
   print "normalizing sequent\n",
   norm_seq Sig Seq Seq',
   norm_ListSeq Sig L L'.
      

type seqListOK list ty -> list signat -> list sequent -> int -> o.

seqListOK LTy Sig nil 0.
seqListOK LTy Sig [Seq | L] N :-
   if (seqOK LTy Sig Seq)
      (
        seqListOK LTy Sig L N',
        N is N' + 1)
      (
        Seq = (seq Nm _ _ _), 
        print "Invalid sequent: ", print Nm, print ".\n",
        fail
      ).
           
% ------ translating parser-generated abstract syntax tree to 
% internal representation

type absyn_to_seq list linc_seqt -> list sequent -> o.
type linc_makeHyp list linc_seqt -> list signat -> list tm -> o.
type absyn_to_seq1 linc_seqt -> sequent -> o.
type absyn_to_def linc_defRoot -> defn -> o.

absyn_to_seq1 (linc_seqRoot Name L Goal) (seq Name LSig LHyp Goal) :-
   linc_makeHyp L LSig LHyp.

absyn_to_seq nil nil.
absyn_to_seq [H | L] [G | R] :-
   absyn_to_seq1 H G, absyn_to_seq L R.
      
linc_makeHyp nil nil nil.
linc_makeHyp [(linc_seqSig LSig1)|L] LSig LHyp :-
   linc_makeHyp L LSig2 LHyp,
   append LSig1 LSig2 LSig.
linc_makeHyp [(linc_seqHyp LHyp1)|L] LSig LHyp :-
   linc_makeHyp L LSig LHyp2,
   append LHyp1 LHyp2 LHyp.      

absyn_to_def (linc_definition Name L) (def Name Kind Sig Cls) :-
   linc_makeDef L Kind Sig Cls.

type linc_makeDef list linc_def -> list ty -> list signat -> list defclause -> o.
linc_makeDef nil nil nil nil.
linc_makeDef [(linc_kindDec LTy)|L] LK Sig Cls :-
   linc_makeDef L LK1 Sig Cls,
   append LTy LK1 LK.
linc_makeDef [(linc_typeDec LTy)|L] LK Sig Cls :-
   linc_makeDef L LK Sig1 Cls, 
   append LTy Sig1 Sig.
linc_makeDef [(linc_defDec DefCl)|L] LK Sig [DefCl|Cls] :-
   linc_makeDef L LK Sig Cls.
         
   

% -- focusing proofs ---
type left_async tm -> o.
type right_async tm -> o.

left_async lfalse.
left_async ltrue.
left_async (land @ A @ B).
left_async (lor @ A @ B).
left_async (lexists @ L).

right_async ltrue.
right_async (land @ A @ B).
right_async (limp @ A @ B).
right_async (lall @ L).

type left_async_forms list tm -> o.

left_async_forms nil.
left_async_forms [T|L] :- 
   left_async T, left_async_forms L.
   

type focus2 list sequent -> list sequent -> o.

focus2 nil nil.
focus2 [S|LSeq1] LSeq2 :- 
   S = (seq Nm Sig Hyp G),
   if (right_async G) 
      (
         infer_right G S L,
         append L LSeq1 LSeq3,
         focus2 LSeq3 LSeq2
      )
      (
         if (remove H Hyp Hyp1, left_async H, !)
            (
              infer_left H (seq Nm Sig Hyp1 G) L2, 
              append L2 LSeq1 LSeq4,
              focus2 LSeq4 LSeq2
            )
            (
              focus2 LSeq1 LSeq5,
              LSeq2 = [S | LSeq5]
            )
      ).

type infer_right tm -> sequent -> list sequent -> o.

infer_right ltrue Seq nil :-
   Seq = (seq Nm _ _ _),
   print "Applying top-rule to sequent ", print Nm,
   print "\n".
   
infer_right (land @ A @ B) Seq LSeq :-
   Seq = (seq Nm _ _ _),
   print "Applying and-right-rule to sequent ", print Nm,
   print "\n",
   inf inf_andr (land @ A @ B) Seq LSeq _.
infer_right (limp @ A @ B) Seq LSeq :-
   Seq = (seq Nm _ _ _),
   print "Applying imp-right-rule to sequent ", print Nm,
   print "\n",
   inf inf_impr (limp @ A @ B) Seq LSeq _.      
infer_right (lall @ L) Seq LSeq :-
   Seq = (seq Nm _ _ _),
   print "Applying pi-right-rule to sequent ", print Nm,
   print "\n",
   inf inf_pir (lall @ L) Seq LSeq _.

type infer_left tm -> sequent -> list sequent -> o.
      
infer_left lfalse Seq nil :-
   Seq = (seq Nm _ _ _),
   print "Applying bot-rule to sequent ", print Nm,
   print "\n".
infer_left ltrue Seq [Seq] :-
   Seq = (seq Nm _ _ _),
   print "Applying weakening-left-rule to sequent ", print Nm,
   print "\n".

infer_left (land @ A @ B) Seq LSeq :-    
   Seq = (seq Nm _ _ _),
   print "Applying and-left-rule to sequent ", print Nm,
   print "\n",
   inf inf_andl (land @ A @ B) Seq LSeq _.
infer_left (lor @ A @ B) Seq LSeq :-
   Seq = (seq Nm _ _ _),
   print "Applying or-left-rule to sequent ", print Nm,
   print "\n",
   inf inf_orl (lor @ A @ B) Seq LSeq _.
infer_left (lexists @ L) Seq LSeq :-
   Seq = (seq Nm _ _ _),
   print "Applying sigma-left-rule to sequent ", print Nm,
   print "\n",
   inf inf_sigmal (lexists @ L) Seq LSeq _.

% optional : defL and eqL
infer_left (leq @ S @ T) Seq LSeq :-
   Seq = (seq Nm _ _ _),
   print "Applying equality-left to sequent ", print Nm,
   print "\n",
   inf inf_eql (leq @ S @ T) Seq LSeq _.
   
%infer_left A Seq LSeq :-
%   atomic A _, !,
%   Seq = (seq Nm _ _ _),
%   print "Applying definition-left to sequent ", print Nm,
%   print "\n",
%   inf inf_eql A Seq LSeq _.
   
type count_sequents list sequent -> int -> o.

count_sequents nil 0.
count_sequents [_|L] N :-
   count_sequents L N1, N is N1 + 1.
   
focus Seq LSeq N :-
   focus2 [Seq] LSeq,
   count_sequents LSeq N.
 
focus_eql Seq LSeq N :-
   (pi S\ pi T\ left_async (leq @ S @ T)) 
   => (focus2 [Seq] LSeq, count_sequents LSeq N).
   
% --- input term from file

input_fterm File T :-
   open_in File In,
   ((pi It\ (pi LA\ (pi (LN\ (yyinitial LN It [eols] :- 
	 	   lookahead It LA, eof_symbol LA, !))))) => (eof_symbol "\n" => yyinitial 0 In Tokens)),
   close_in In, !,
   startparse (linc_term St) [bols] Tokens D Str,
   D = (linc_term T), !.  
   
inff File inf_sigmar G (seq Nm LFv Hyp G) LSeq 1 :-
   G = (lexists @ L),
   Nm1 is Nm ^ "1",
   input_fterm File T,
   current_def (def DefNm Kind Sig LCls),
   Seq' = (seq Nm1 LFv Hyp (L @ T)),
   seqOK Kind Sig Seq', norm_seq Sig Seq' Seq1,
   LSeq = [Seq1].
   
inff File inf_coindr G (seq Nm LFv Hyp G) LSeq 2 :-
   Nm1 is Nm ^ "1", Nm2 is Nm ^ "2",
   atomic G [A|Args],
   current_def (def DefNm Kind Sig LCls),
   member (defcl coindcl A B) LCls, !,
   input_fterm File I,  % input the induction invariant
   append LFv Sig Sig', 
   typeof Kind Sig' I Ty, typeof Kind Sig' A Ty,
   subst_invariant A I B B', 
   new_var_args 0 Args Kind Sig' NewSig NewVars,  % create new eigenvariables 
   apply_atom I Args C,   
   apply_atom B' NewVars BI,
   apply_atom I NewVars SI,
   eval Sig' C D,         
   append NewSig Sig Sig2,
   eval Sig2 BI BInorm,
   eval Sig2 SI SInorm,
   LSeq = [seq Nm1 NewSig [SInorm] BInorm, seq Nm1 LFv Hyp D].   
   
inff File inf_pil (lall @ L) (seq Nm LFv Hyp G) LSeq 1 :-
   Nm1 is Nm ^ "1",
   input_fterm File T,
   current_def (def DefNm Kind Sig LCls),
   Seq' = (seq Nm1 LFv [L @ T|Hyp] G),
   seqOK Kind Sig Seq', norm_seq Sig Seq' Seq1,
   LSeq = [Seq1].   
   
inff File inf_indl P (seq Nm LFv Hyp G) LSeq 2 :- 
   Nm1 is Nm ^ "1", Nm2 is Nm ^ "2",
   atomic P [A|Args],
   current_def (def DefNm Kind Sig LCls),
   member (defcl indcl A B) LCls, !,
   input_fterm File I,  % input the induction invariant
   append LFv Sig Sig', 
   typeof Kind Sig' I Ty, typeof Kind Sig' A Ty,
   subst_invariant A I B B', 
   new_var_args 0 Args Kind Sig' NewSig NewVars,  % create new eigenvariables 
   apply_atom I Args C,   
   apply_atom B' NewVars BI,
   apply_atom I NewVars SI,
   eval Sig' C D,         
   append NewSig Sig Sig2,
   eval Sig2 BI BInorm,
   eval Sig2 SI SInorm,
   LSeq = [seq Nm1 NewSig [BInorm] SInorm, seq Nm1 LFv [D|Hyp] G].   