% Module unif
%
% Copyright (C) 2003 Alwen Tiu
%
%   This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or (at
% your option) any later version.
% 
%   This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software Foundation,
% Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


% Module:  Huet's Unification Algorithm
% Author:  Alwen Tiu
% Date:    11/30/01, 8/26/03

% Revision:
% - 8/31/03 : add a heuristic to umatch, in the case where the pair of flex-rigid
%             is (X, e) where X is a free variable. Occur check is performed on e.
%             add a heuristic to flex-flex case (see unify_two_fvars).


module unif.

accum_sig stt.
accumulate stt, list, control.

% list of declared base types.
type cur_kinds list ty -> o.

% bound variables
type bvar tm -> o.

% --substitution------------------------------------------------------

subst Sig nil X Y :- linc_copy X Y.
subst Sig [(pair X Y) | L] T U :-
   (linc_copy X Y :- !) => subst Sig L T U.


% substitution composition.
% (f o g)(t) = f(g(t)).

type compose list signat -> list spair -> list spair -> list spair -> o.
type rsubst list signat -> list spair -> list spair -> list spair -> o.
type rem_red list spair -> list spair -> list spair -> o.

% rsubst S L K
% Apply substitution S to the range of substitution L.

rsubst Sig S nil nil.
rsubst Sig S [(pair U V) | L] K :-
    subst Sig S V W, rsubst Sig S L K',
    if (U = W) (K = K') (K = [(pair U W) | K']).

% rem_red S L K.
% remove substitution pair in S that is already in K.

rem_red nil L L.
rem_red [(pair X Y) | S] L K :-
   member (pair X W) L, !, rem_red S L K.
rem_red [P | S] L [P | K] :-
   rem_red S L K.

compose Sig L K M :-
   rsubst Sig L K K',
   rem_red L K' M.

% --------------------------------------------------------------------

type rigid tm -> o.
type flex  list signat -> tm -> o.

% Description:
%   Check wheter a normal term has a rigid or flexible head.
%   Head of a normal term is defined as follows:
%       Suppose M = lam x1\ ... lam xn\ x @ .....
%       where x is either a constant or a variable (free or bounded).
%       The head of M is x.
%       It is called flexible if x is a free variable, otherwise rigid.
%

rigid (lam Ty M) :-
  !, pi x\ (rigid x) => rigid (M x).
rigid (M @ N) :-
  !, rigid M.
rigid C :-
  constant (sg C _).

flex Sig (lam Ty M) :-
  !,pi x\ flex Sig (M x).
flex Sig (M @ N) :-
  !,flex Sig M.
flex Sig V :- member (sg V _) Sig, !, not (constant (sg V _)).

% --

type check_heading tm -> tm -> o.

% Description : check_heading M N.
%   Check if the two normal terms M and N has the same heading.
%   It applies only to rigid-rigid pairs. 
% Ex: check_heading (lam x\ lam y\ y @ x) (lam z\ lam k\ k @ z)
%     yes.
%

check_heading (lam Ty M) (lam Ty N) :-
   !, pi x\ check_heading x x => check_heading (M x) (N x).
check_heading (M @ N) (P @ Q) :-
   !, check_heading M P.
check_heading C C :- constant (sg C _).

% --

type getbs1 ty -> ty -> o.
type get_bs_types list signat -> list ty -> list ty -> o.

% Description:
%   getbs1 T1 T2.
%   Given T1 = a1 --> a2 --> .... --> b, getbs1 return T2 = b, where b
%   is a base type.
%   get_bs_types iterates getbs1 on a list of types, and gives a result of
%   list of distinct base types.
% 

getbs1 (U --> V) T :-
   !, getbs1 V T.
getbs1 X X.
        
get_bs_types nil L L.
get_bs_types [(sg M T) | L]  K H :-
   getbs1 T T', 
   (if (member T' K)
       (get_bs_types L K H)
       (get_bs_types L (T'::K) H)
   ).

% --

type getfv1 list signat -> tm -> list signat -> list signat -> o.
type get_fvars list signat -> list spair -> list signat -> list signat -> o.

% Description:
%   getfv1 Sig T Sig1 Sig2.
%   getfv1 returns a list of free variables occuring in term T, along with their types.
%   The return value is saved in Sig2, and Sig1 acts as an accumulator.
%   get_fvars apply getfv1 to a list of pairs of terms.

getfv1 Sig (lam Ty M) L K :-
   pi x\ (pi S\ pi L'\  getfv1 S x L' L') =>  getfv1 Sig (M x) L K.
getfv1 Sig (M @ N) L K :-
   getfv1 Sig M L L', getfv1 Sig N L' K.
getfv1 Sig V L L :-
   member (sg V T) Sig, member (sg V T) L.
getfv1 Sig V L [(sg V T) | L] :-
   member (sg V T) Sig, not (member (sg V T) L).
getfv1 Sig C L L :-
   constant (sg C _).

get_fvars Sig nil L L.
get_fvars Sig [(pair S T) | L] M K :-
   getfv1 Sig S M M1, getfv1 Sig T M1 M2, 
   get_fvars Sig L M2 K.

% --

type create_new_fvars int -> list signat -> list ty -> list signat -> o.

% Description:
%   create_new_fvars N SigIn L SigOut.
%   Create new variables, one for each type in L. The result is saved in SigOut.
%  

create_new_fvars N Sig nil nil.
create_new_fvars N Sig (T::L) [(sg FVar T) | K] :-
   get_next_fvar Sig N FVar, 
   N' is N + 1, create_new_fvars N' [(sg FVar T)|Sig] L K.

% --

type unify_fvar1 ty -> list signat -> tm -> o.
type unify_fvars list signat -> list signat -> list spair -> o.

% Description: unify_fvar1 Ty Sig Var. 
%   These predicates produces a unifier for a node that consists only of flex-flex pairs.
%   unify_fvar1 applies to a variable, while unify_fvars apply to a list of variables.
%   Each free variable F of type a base type T0 is replaced by a (new) variable in the list
%   of variables (Sig), and composite types are replaced by:
%      T = (a1 --> a2 --> ... --> an --> b), b is a base type,
%      then take the term lam w1\ ... lam wn\ h, where wi is of type ai, and h (of type b) 
%      is different from w1 ... wn.

unify_fvar1 (U --> V) Sig (lam U x\ M) :-
   pi x\ (pi S\ unify_fvar1 U S x) => unify_fvar1 V Sig M.
unify_fvar1 T Sig M :-
   member (sg M T) Sig.

unify_fvars nil Sig nil.
unify_fvars [(sg V T) | L] Sig [(pair V M) | K] :-
   unify_fvar1 T Sig M,
   unify_fvars L Sig K.

% --

type case_r_r list signat -> list signat -> spair -> 
                     list spair -> node -> list spair -> o.

% Description:
%   Subcase of Step 1 in SIMPL (see simpl predicate below).
%   case_r_r SigIn SigOut (pair X Y) L Node Subs.
%   Given two rigid terms X and Y, it compares their heads (using check_heading).
%   If they are the same then it decomposes the pair in the following way:
%       X = lam u1\ ... lam un\ h1 @ e1 ... @ ep
%       Y = lam v1\ ... lam vn\ h2 @ f1 ... @ fp
%     then take N' = [(pair (lam u1\ ... lam un\ ei) (lam v1\ .. lam vn\ fi) )], for 1 <= i <= p.
%   The decomposed pairs are added to current node.
 
type getlastarg tm -> tm -> tm -> o.
type arglist tm -> list tm -> o.
type pairlist list tm -> list tm -> list spair -> o.

case_r_r Sig Sig' (pair X Y) L N Subs :-
   check_heading X Y, !, 
   arglist X L1, arglist Y L2, 
   pairlist L1 L2 L3, append L3 L K,
   simpl Sig Sig' (n_node K) N Subs.
case_r_r Sig Sig (pair X Y) L fail_node nil.

% --

type pairlist list tm -> list tm -> list spair -> o.

% pair two lists of terms.
%
pairlist [X] [Y] nil.
pairlist [X|L] [Y|K] [(pair X Y) | M] :-
   pairlist L K M.

% --

% Description:
%    getlastarg T M N.
%    Given a term T = lam x1\ .. lam xn\ P @ Q, getlastarg produces
%    the term M = lam x1\ ... lam xn\ P 
%    and the term N = lam x1\ ... lam xn\ Q.
%    If the term consist only of head, like lam x1\ .. lam xn\ xi,
%    then M = T.
%
getlastarg (M @ N) M N :- !.
getlastarg (lam T M) (lam T P) (lam T Q) :-
   !, pi x\ getlastarg (M x) (P x) (Q x).
getlastarg X X X.

% Description:
%   arglist T L.
%   Decompose a term T = lam x1\ ... lam xn\ p1 @ ... @ pm into
%      [ lam x1\ .. lam xn\ pi ], where 1 <= i <= m, (in decreasing order).
%
arglist M [M] :-
   getlastarg M R L, M = R, !.
arglist M [N|L] :-
   getlastarg M R N, arglist R L.

% ---------------------------------------------------------

type simpl list signat -> list signat -> node -> node -> list spair -> o.

% Description:
%   Procedure SIMPL. See [Huet75] for details.
%   simpl SigIn SigOut NodeIn NodeOut Subs.
%   SigIn is the original signature, SigOut is the output signature, because
%   unification will probably introduces new variables. 
%   NodeIn, NodeOut : input-output nodes.
%   Subs is the unifier, if it exists.

simpl Sig Sig fail_node fail_node nil.
simpl Sig Sig succ_node succ_node nil.
simpl Sig Sig (n_node nil) succ_node nil.
simpl Sig Sig flex_node flex_node nil.

% Step 1. Rigid-rigid case. Decompose all the matching pairs.
%         Actual decomposition is done in case_r_r.
simpl Sig Sig' (n_node [H|L]) N Subs :-
   remove (pair X Y) [H|L] K,
   rigid X, rigid Y, !, 
   case_r_r Sig Sig' (pair X Y) K N Subs.  

% Step 2. Flip all the rigid-flex pairs to flex-rigid pairs.
simpl Sig Sig' (n_node [H|L]) N Subs :-
   remove (pair X Y) [H|L] K,
   rigid X, flex Sig Y, !,
   simpl Sig Sig' (n_node [(pair Y X) | K]) N Subs.

% Step 3.a. If there is a flex-rigid pair in the node, return the node.
simpl Sig Sig (n_node [H|L]) (n_node [H|L]) nil :-
   remove (pair X Y) [H|L] K,
   rigid Y, !.

%      3.b. Otherwise, all the pairs are flex-flex, find the unifier.
simpl Sig Sig' (n_node [H|L]) Node Subs :-
   if (unify_two_fvar Sig [H|L] Subs)     % try first if they are variable pairs
      (Sig' = Sig, Node = succ_node)
      (
		 Node = flex_node,
         get_fvars Sig [H|L] nil SigFVars,     % get all free variables occuring in node
         get_bs_types SigFVars nil LTypes,     % get the base types 
         create_new_fvars 0 Sig LTypes NewSig, % create new variables for base types
         append Sig NewSig Sig',               % update signatures
         unify_fvars SigFVars NewSig Subs     % unify terms in the node.
      ).

% --

% unify a node which consists only of variable-variable pairs

type unify_two_fvar list signat -> list spair -> list spair -> o.

unify_two_fvar Sig nil nil.
unify_two_fvar Sig [(pair T T) | L] Subs :-
    !, unify_two_fvar Sig L Subs.
unify_two_fvar Sig [(pair (fvar X) T) | L] Subs :-
    !, not(occur (fvar X) T), 
	subst_list Sig [(pair (fvar X) T)] L L',
	unify_two_fvar Sig L' Subs',
	compose Sig Subs' [pair (fvar X) T] Subs.

unify_two_fvar Sig [(pair T (fvar X)) | L] Subs :-
    !, not(occur (fvar X) T),
	subst_list Sig [(pair (fvar X) T)] L L',
	unify_two_fvar Sig L' Subs',
	compose Sig Subs' [pair (fvar X) T] Subs.
		
% count number of abstraction and arguments in a
% normal term.
type count_arg tm -> int -> int -> o.

count_arg (lam T M) A B :-
   !, pi x\ count_arg (M x) A' B,
   A is A' + 1.
count_arg (M @ N) A B :-
   !, count_arg M A B', B is B' + 1.
count_arg X 0 0.

% --

type get_head_idx tm -> int -> int -> o.
type var_idx tm -> int -> o.

% Description:
%   if the head of a normal term is a bound variable, this predicate will
%   return the index of the bound variable or 0 if it's constant.
%   ex:  get_head (lam x1\ lam x2\ x2) 0 N.
%        N = 2.

var_idx C 0 :- constant (sg C _).

get_head_idx (lam T M) SIdx Idx :-
  I is SIdx + 1, 
  pi x\ (var_idx x SIdx) => get_head_idx (M x) I Idx.
get_head_idx (M @ N) SIdx Idx :-
  get_head_idx M SIdx Idx.
get_head_idx X SIdx Idx :-
  var_idx X Idx.

% --

type get_head int -> tm -> tm -> o.

% Description:
%   get_head N T1 T2.
%   T1 is a normal term with N' abstractions.
%   get_head produces term T2 with (N' - N) abstractions and
%   only the head of term T1 is included in T2.
% ex : get_head 1 (lam x\ lam y\ y @ x) T2. 
%      T2 = lam y\ y.

get_head 0 (lam T M) (lam T N) :-
  pi x\ get_head 0 x x => get_head 0 (M x) (N x).
get_head 0 (M @ N) P :-
  get_head 0 M P.
get_head 0 C C :-
  constant (sg C _).
get_head I (lam T M) N :-
  I > 0, I' is I - 1, 
  pi x\ get_head I' (M x) N.

% -- 

type get_head_var list signat -> tm -> tm -> o.

% Description: get_head_var Sig M V.
%   Get the head variable of a normal term M.
% ex:  get_head_var (lam x\ a @ x) V.
%      V = a.

get_head_var Sig (lam T M) V :-
  pi x\ get_head_var Sig (M x) V.
get_head_var Sig (M @ N) V :-
  get_head_var Sig M V.
get_head_var Sig X X :-
  member (sg X T) Sig.

% --
type list2tm tm -> list tm -> tm -> o.

% Description : 
%   list2tm H L T.
%   Apply the term H to the list of terms in L (in the reverse order).
%
% ex: list2tm a [b, c, d] T.
%     T = a @ d @ c @ b.

list2tm M nil M.
list2tm H [X|L] (R @ X) :- list2tm H L R. 

% --

type apply_head list signat -> list signat -> int -> list tm -> tm -> tm -> o.

% Description :
%   apply_head SigIn SigOut N L H T.
%   Produce a term with the head H, with N arguments.
%   Each arguments has the form Hi @ E1 @ ... @ En, where E1..En are the arguments
%   contained in list L. New variables introduced are placed in SigOut.
%
%   ex: apply_head nil Sig 2 [a, b] h T.
%       T = h @ ((fv 1) @ b @ a) ( (fv 2) @ b @ a)
%       Sig = [(sg (fv 1) T1), (sg (fv 2) T2)].
  
apply_head Sig Sig 0 L T T.
apply_head SigIn SigOut N L R T :-
   N > 0, N' is N - 1,
   get_next_fvar SigIn 0 FVar,
   S' = [(sg FVar Ty) | SigIn],        % update the signature
   list2tm FVar L E,                 % apply the new variable to the list of args L.
   apply_head S' SigOut N' L (R @ E) T.

% -- 

type imitate_term list signat -> list signat -> int -> int -> list tm -> tm -> tm -> o.

% Description:
%   imitate_term SigIn SigOut P1 P2 L M N.
%   Take a term M, and apply P1 number of abstractions to M.
%   The head of M is also applied to P2 arguments, using apply_head (see apply_head for 
%   explanation). SigIn and SigOut are, as usual, used for updating signatures.
%   L is the initial list of argument, and it should be nil when it's called from
%   other predicates.
%
% ex: imitate_term nil Sig 2 1 nil and M.
%     M = lam x\ lam y\ (and @ ( (fv 1) @ x @ y)).
%     Sig = [(sg (fv 1) T1)].
%
% ex: imitate_term nil Sig 2 1 nil (lam x\ lam y\ x) M.
%     M = lam x'\ lam y'\ lam x\ lam y\ (x @ ( (fv 1) @ x' @ y' @ x @ y))
%     Sig = [(sg (fv 1) T1)].

imitate_term SigIn SigOut P1 P2 L M (lam Ty N) :-
   P1 > 0, P1' is P1 - 1,
   pi x\ imitate_term SigIn SigOut P1' P2 (x::L) M (N x).
imitate_term SigIn SigOut 0 P2 L (lam T M) (lam T N) :-
   pi x\ bvar x => imitate_term SigIn SigOut 0 P2 (x::L) (M x) (N x).
imitate_term SigIn SigOut 0 P2 L C T :-
   (constant (sg C _); bvar C), 
    apply_head SigIn SigOut P2 L C T.


% --
type umatch, imitate, project list signat -> list signat -> spair -> list spair -> o.

% Description:
%   Imitation rule [Huet75].
%

imitate SigIn SigOut (pair X Y) Subs :-
  count_arg X N1 P1, count_arg Y N2 P2,
  get_head_var SigIn X FV,
  if (N2 = N1)  
     ( % then
      get_head N1 Y HY, 
      imitate_term SigIn Sig P1 P2 nil HY HX,
      cur_kinds Kn, 
      if (typeof Kn Sig FV Ty, typeof Kn Sig HX Ty)
         (Subs = [(pair FV HX)], SigOut = Sig)
         (Subs = nil, SigOut = SigIn)
     ) % else 
     (
      SigOut = SigIn, Subs = nil
     ).

% --

type get_type_at int -> ty -> ty -> o.
type get_type_art ty -> int -> o.

% Description:
%   get_type_at N T1 T2,
%     get the N-th argument of type T1, the result is T2.
%   get_type_art T N, get the arity of type T.
% 

get_type_art (U --> V) N :-
   !, get_type_art V N', N is N' + 1.
get_type_art X 0.

get_type_at 1 (U --> V) U.
get_type_at N (U --> V) T :-
   N > 1, N' is N - 1,
   get_type_at N' V T.

% --  

type project_term list signat -> list signat -> int -> int -> int -> list tm -> tm -> o.
type prj_head tm -> o.

% Description:
%   project_term SigIn SigOut I K M L Tm.
%   Create a term : lam w_1 ... w_K\ w_I,
%   The predicate prj_head is to store w_I (since it could be a bind variable).

project_term SigIn SigOut 0 0 M L T :-
   prj_head H, apply_head SigIn SigOut M L H T.
project_term SigIn SigOut 1 K M L (lam Ty Tm) :-
   K > 0, K' is K - 1,
   pi x\ prj_head x => project_term SigIn SigOut 0 K' M (x::L) (Tm x).
project_term SigIn SigOut 0 K M L (lam Ty Tm) :-
   K > 0, K' is K - 1,
   pi x\ project_term SigIn SigOut 0 K' M (x::L) (Tm x).
project_term SigIn SigOut I K M L (lam Ty Tm) :-
   I > 1, K > 0, I' is I - 1, K' is K - 1,
   pi x\ project_term SigIn SigOut I' K' M (x::L) (Tm x).

% --

type itr_project list signat -> list signat -> int -> int -> tm -> list spair -> o.

% Description:
%   iterate project_term.

itr_project Sig Sig 0 K FV nil. 
itr_project SigIn SigOut 1 K FV Subs :-
   member (sg FV T) SigIn,
   get_type_at 1 T A,
   getbs1 T B,
   if (getbs1 A B)
      (
        get_type_art T Art1, get_type_art A Art2,
        M is Art2 - (Art1 - K),
        if (M >= 0)
           ( 
             project_term SigIn Sig 1 K M nil Tm,
             cur_kinds Kn, 
             if (typeof Kn Sig Tm T)
                (Subs = [(pair FV Tm)], SigOut = Sig)
                (Subs = nil, SigOut = SigIn)
           )
           (Subs = nil, SigOut = SigIn)
      )
      (Subs = nil, SigOut = SigIn).

itr_project SigIn SigOut I K FV Subs :-
   I > 1, I' is I - 1,
   member (sg FV T) SigIn,
   get_type_at I T A,
   getbs1 T B,
   if (getbs1 A B)
      (
        get_type_art T Art1, get_type_art A Art2,
        M is Art2 - (Art1 - K),
        if (M >= 0)
           (
             project_term SigIn Sig I K M nil Tm,
             cur_kinds Kinds, 
             if (typeof Kinds Sig Tm T)
                (
                  itr_project Sig SigOut I' K FV Subs',
                  Subs = [(pair FV Tm) | Subs']
                )       
                ( itr_project SigIn SigOut I' K FV Subs )
           )
           ( itr_project SigIn SigOut I' K FV Subs )
        )
       (itr_project SigIn SigOut I' K FV Subs ).

% --
% Description
%   projection rule [Huet75].
%

project SigIn SigOut (pair X Y) Subs :-
   count_arg X N1 P1, count_arg Y N2 P2,
   get_head_var SigIn X FV,
   itr_project SigIn SigOut P1 P1 FV Subs.


% --

% Description: occur check

type occur tm -> tm -> o.

occur X X :- !.
occur X (M @ N) :- (occur X M, !) ; (occur X N).
occur X (lam Ty L) :- pi x\ occur X (L x).

% Description:
%   matching algorithm. See [Huet75].
%

umatch SigIn SigOut (pair X Y) Subs :-
%  print "\n[umatch] subst pair: (", print_lterm std_out 0 X, 
%  print ", ", print_lterm std_out 0 Y, print "\n",
  count_arg X N1 P1, count_arg Y N2 P2,  
  get_head_idx Y 1 N3,
  if (N1 > N2)                        % if 
     (SigOut = SigIn, Subs = nil)     % then
     (
       if (N1 = 0 & not(occur X Y))
       (
          Subs = [(pair X Y)], SigOut = SigIn
       )
       (
         if (imitate SigIn SigImt (pair X Y) S1)
            (Subs1 = S1, Sig1 = SigImt)
            (Subs1 = nil, Sig1 = SigIn), !,
         if (project Sig1 SigPrj (pair X Y) S2)
            (Subs2 = S2, Sig2 = SigPrj)
            (Subs2 = nil, Sig2 = Sig1), !,
         append Subs1 Subs2 Subs,
         SigOut = Sig2
       )
     ).

% --

% Description:
%   Higher order unification. See [Huet75] for details.
%

type matchtree list signat -> list signat -> node -> node -> 
               list spair -> list (list spair) -> o.
type matchtreelst list signat -> list signat -> list node -> node ->
               list (list spair) -> list (list spair) -> o.

type subst_list list signat -> list spair -> list spair -> list spair -> o.
type subst_node list signat -> list spair -> list spair -> list node -> o.
type eval_node  list signat -> list spair -> list spair -> o.
type compose_list list signat -> list spair -> list spair -> list (list spair) -> o.

% Reduce all the term in a node to beta-normal form.

eval_node Sig nil nil.
eval_node Sig [(pair X Y) | L] [(pair S T) | K] :-
   eval Sig X S, eval Sig Y T, 
   eval_node Sig L K.

% apply substitution to a list of pair of terms.

subst_list Sig Subs nil nil.
subst_list Sig Subs [(pair X Y) | L] [(pair P Q) | K] :-
   subst Sig Subs X P, subst Sig Subs Y Q,
   subst_list Sig Subs L K.

% apply substitution to a node.

subst_node Sig nil LPair nil.
subst_node Sig [S | LSubs] LPair [(n_node N) | LNode] :-
    subst_list Sig [S] LPair N,
    subst_node Sig LSubs LPair LNode.

% compose all the substitution in a list of substitutions.

compose_list Sig nil SubIn nil.
compose_list Sig [S | L] SubIn [T | K] :-
   compose Sig [S] SubIn T,
   compose_list Sig L SubIn K. 

% printing node

type print_node node -> o.

print_node fail_node :- print "fail-node\n".
print_node succ_node :- print "success-node\n".
print_node flex_node :- print "flex-node\n".
print_node (n_node nil) :- print " \n".
print_node (n_node [(pair X Y) | L]) :-
   term_to_string X StrX, term_to_string Y StrY,
   print "[ ", print StrX, print " -- ", print StrY, print " ]\n",
   print_node (n_node L).

% --

type expand_tm tm -> ty -> tm -> list tm -> o.

% Description:
%   replace a term C : a1 --> a2 --> ... --> an -> b
%   with lam x1\ ... lam xn\ C @ x1 @ ... @ xn.

expand_tm C (U --> V) (lam Ty M) Args :-
  !, pi x\ expand_tm C V (M x) (x::Args).
expand_tm C Ty M Args :-
  list2tm C Args M.

% --

type eta_expand list signat -> tm -> tm -> o.

% Description:
%   replaces a term lam u1\ ... lam un\ C @ e1 @ .. @ ep
%   where C : a1 --> .. --> ap --> .. --> aq --> b,
%   with the term :
%    lam u1 .. lam un\ lam w1 .. lam w(q-p)\ C @ e1 .. @ ep @ w1 @ .. @ w(q-p).
%
 
eta_expand Sig (lam Ty M) (lam Ty N) :-   
  cur_kinds Kinds, 
  (pi x\ (bvar x, (pi S\ typeof Kinds S x Ty)) => eta_expand Sig (M x) (N x)).
eta_expand Sig (M @ N) M' :-
  cur_kinds Kinds,
  typeof Kinds Sig (M @ N) Ty, expand_tm (M @ N) Ty M' nil.
eta_expand Sig C M :-
  cur_kinds Kinds,
  constant (sg C _), !, typeof Kinds Sig C Ty, expand_tm C Ty M nil.
eta_expand Sig C M :-
  member (sg C Ty) Sig, expand_tm C Ty M nil.
eta_expand Sig C M  :-
  cur_kinds Kinds,
  bvar C, typeof Kinds Sig C Ty, expand_tm C Ty M nil.

% --

type eta_expand_node list signat -> list spair -> list spair -> o.

eta_expand_node Sig nil nil.
eta_expand_node Sig [(pair X Y) | L] [(pair X' Y') | K] :-
  eta_expand Sig X X', eta_expand Sig Y Y',
  eta_expand_node Sig L K.

% --
% Description:
%   Generate match tree.
%   SubIn : substitution, LSubOut : List of substitutions.

matchtree SigIn SigOut (n_node L) Node SubIn LSubOut :-
%   print "\n [matchtree] Eval node",
   eval_node SigIn L L1,
   eta_expand_node SigIn L1 L',
%   print "\n [matchtree] Simpl",
%   print "\n  node before : ", term_to_string (n_node L') StrNB, print StrNB,
   simpl SigIn Sig (n_node L') N S1,
%   print "\n   node : ", term_to_string N StrN, print StrN,
   if (N = (n_node [H|K])) 
      (
         member (pair X Y) [H|K], 
         flex Sig X, rigid Y, !,
%         print "\n [matchtree] umatch",
         eval Sig X X1, 
         eval Sig Y Y1,
         eta_expand Sig X1 X2, 
         eta_expand Sig Y1 Y2, 
         umatch Sig Sig1 (pair X2 Y2) S2,
%         print "\n    Subs ", term_to_string S2 StrS2, print StrS2,
         subst_node Sig1 S2 [H|K] LNode,
         compose_list Sig1 S2 SubIn S3,      
         matchtreelst Sig1 SigOut LNode Node S3 LSubOut
      )
      ( 
%         print "\ndodol\n",
         if (N = (n_node nil))
            (Node = succ_node)
            (Node = N),
         SigOut = Sig,
         if (Node = succ_node; Node = flex_node)
            (
%              print "compose\n",
              compose Sig S1 SubIn S4,
              LSubOut = [S4]
            )
            ( LSubOut = nil)
      ).

matchtreelst SigIn SigOut [HNode] Node [SubIn] LSubOut :-
   matchtree SigIn SigOut HNode Node SubIn LSubOut.
matchtreelst SigIn SigOut [HNode, H2 | LNode] Node [SubIn, Sub2 | LSubIn] LSubOut :-
   matchtree SigIn Sig HNode N1 SubIn LS1,
   matchtreelst Sig Sig' [H2|LNode] N2 [Sub2|LSubIn] LS2,
   SigOut = Sig',
   if (N1 = flex_node) (Node = N1) 
      (  if (N1 = succ_node, not (N2 = flex_node)) 
            (Node = N1)
            (Node = N2)
      ),
   append LS1 LS2 LSubOut.               

% --

type unify1 list signat -> list signat -> tm -> tm -> node -> list (list spair) -> o.

% Finally, the main procedure for unification algorithm!

unify1 SigIn SigOut T1 T2 N LSubs :-
%  print "entering unify1\n",
  cur_kinds Kinds,
  if (typeof Kinds SigIn T1 A, typeof Kinds SigIn T2 A)   % check if terms are well-typed.
     (
%       print "entering matchtree\n",
       matchtree SigIn SigOut (n_node [(pair T1 T2)]) N nil LSubs
     ) 
     (N = fail_node, LSubs = nil, SigOut = SigIn).

type unify defn -> list signat -> list signat -> tm -> tm -> node -> list (list spair) -> o.

   
unify Def SigIn SigOut T1 T2 N LSubs :-
   Def = (def _ Kinds Sig _), 
   ( cur_kinds Kinds => 
     unify1 SigIn SigOut T1 T2 N LSubs
    ).
          
% --





